import time

from .display import *
from .input import *
__all__ = [
    'ReturnInterrupt', 'RetryInterrupt', 'call_retriable_func'
]

class ReturnInterrupt(Exception):
    def __init__(self, return_value):
        self.return_value = return_value

class RetryInterrupt(Exception):
    pass

def call_retriable_func(func: callable, max_retries: int = -1, 
                        wait_time_between_retries: float = 0.2,
                        *args, **kwargs):
    """Calls `func` with `args` and `kwargs`, and recalls it whenever
    it raises `RetryInterrupt` up to `max_retries` times.
    
    If `max_retries` is negative, retries it infinitely.\n
    Between each retry, sleeps `wait_time_between_retries` seconds.
    """
    retries_count = 0

    def retries_remaining() -> int:
        return max_retries - retries_count + 1

    while 1:
        try:
            try:
                return func(*args, **kwargs)
            except RetryInterrupt:
                if retries_count == max_retries:
                    return
                
                retries_count += 1
                msg = 'Retrying'

                if wait_time_between_retries > 0:
                    msg += f' in {wait_time_between_retries} seconds'
                    if max_retries > 0:
                        msg += f'. {retries_remaining()} retries remaining'

                    debug(f'{msg}...')
                    time.sleep(wait_time_between_retries)
                else:
                    debug(f'{msg}...')
        except KeyboardInterrupt:
            # No retries happened so no need to ask anything,
            # as user just wants to abort the execution of func()
            if retries_count == 0:
                return

            # Lets ask just in case, user may want to pause the operation 
            # to fix whatever is causing it to fail
            try:
                if ask(f'{retries_remaining()} retries remaining. Abort?', True):
                    return
            except (KeyboardInterrupt, EOFError):
                return



import os

from .display import *
__all__ = [
    'settings_file', 'Setting'
]

settings_file = 'config'

def _init():
    """Creates `settings_file` if it doesn't exist."""

    if not os.path.isfile(settings_file):
        open(settings_file, 'a').close()

def _read_setting(key: str) -> str | None:
    """Reads the value of `key` from `settings_file` and returns it.

    If the setting is not set, returns `None`.
    """
    _init()
    with open(settings_file) as file:
        for line in file:
            if line.startswith(f'{key}='):
                return line.rstrip()[(len(key) + 1):]

    return None

def _write_setting(key: str, value: str):
    """Writes `key` and `value` to `settings_file`."""

    _init()
    with open(settings_file, 'r+') as file:
        data = file.read()
        for line in data.split('\n'):
            if line.startswith(f'{key}='):
                # TODO: Use a better method than this
                data = data.replace(line, f'{key}={value}')

                # Reset pointer and write
                file.seek(0)
                file.truncate(0)
                file.write(data)
                return
        
        # Key not found, so add it
        file.write(f'{key}={value}\n')

class Setting():
    def __init__(self, key: str, value_type, default_value=None):
        self.key = key
        self.value_type = value_type
        self.default_value = default_value

    def read(self):
        """Reads the value of the setting, parses and returns it.
        
        If the value can't be parsed, prints error
        and returns `default_value` or `None`.

        If the setting is not set and `default_value` isn't `None`,
        sets the setting to `default_value` and returns it.

        Otherwise, returns `None`.
        """

        value = _read_setting(self.key)
        if value is None:
            if self.default_value is not None:
                self.write(self.default_value)

            return self.default_value

        try:
            if self.value_type is bool:
                value = value.lower()
                if value == 'true':
                    value = True
                elif value == 'false':
                    value = False
                else: 
                    raise ValueError('Not a boolean')
            elif self.value_type is int:
                value = int(value)
            elif self.value_type is float:
                value = float(value)
            elif self.value_type is str:
                pass
            else:
                raise ValueError(f'Unknown value type "{self.value_type}"')
        except ValueError:
            error(f'[Setting "{self.key}"] Invalid value "{self.value_type}". ' +
                  f'Returning default value "{self.default_value}"')

            value = self.default_value

        return value

    def write(self, value):
        """Sets the setting's value to `value`.
        
        Returns `True` if successful.
        """

        if not type(value) is self.value_type:
            error(f'[Setting "{self.key}"] New value "{value}" ("{type(value)}") ' +
                  f'doesn\'t match the value type "{self.value_type}"')
            return False

        _write_setting(self.key, value)
        return True


import http.client
import socket
import time
import re
import requests

from .display import *
from .flow import *
__all__ = [
    'FIDDLER_PROXY', 'proxy', 'IP_PATTERN_STR',
    'IP_PATTERN', 'URL_PATTERN', 'Host',
    'request_get','request_post', 'socket_connection'
]

FIDDLER_PROXY = { 
    'http'  : 'http://127.0.0.1:8888',
    'https' : 'https://127.0.0.1:8888'
}
# Set this to fiddler_proxy to make the traffic go through Fiddler
proxy: dict[str, str]|None = None

# 192.168.0.1
IP_PATTERN_STR = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'
IP_PATTERN = re.compile(f'^{IP_PATTERN_STR}$')
# 192.168.0.1
# 192.168.0.1:80
# https://192.168.0.1
# http://192.168.1.1:4568/asd/whatever
# group 1 is the full base URL | group 3 is scheme | group 4 is IP/URL
# group 5 is IP | group 6 is URL | group 9 is port
# group 11 is anything after the slash
URL_PATTERN = re.compile(
    r'(^((https?):\/\/)?((' + IP_PATTERN_STR + r')|((\d|\w|\.)+))(\:(\d+))?)(\/(.+)?)?$'
)
class Host:
    def __init__(self, base_url: str = '') -> None:
        self._scheme: int = ''
        self._domain: str = ''
        self._ip: str = ''
        self._port: int = -1

        if len(base_url) > 0:
            self.base_url = base_url

    @property
    def scheme(self) -> str | None:
        """The scheme of the host. If not set, this value is `None`."""
        if len(self._scheme) == 0:
            return None

        return self._scheme
    
    @scheme.setter
    def scheme(self, value: str|None):
        if value is None:
            self._scheme = ''
            return

        value = value.lower()
        if not value in ['http', 'https']:
            error(f'Host: Invalid scheme "{value}"')
            return

        self._scheme = value

    @property
    def ip(self) -> str | None:
        """The IP address of the host. If not set, this value is `None`."""
        if len(self._ip) == 0:
            return None

        return self._ip
    
    @ip.setter
    def ip(self, value: str|None):
        if value is None:
            self._ip = ''
            return

        match = IP_PATTERN.match(value)
        if match is None:
            error(f'Host: Invalid IP "{value}"')
            return

        self._ip = match.group(0)

    @property
    def port(self) -> int | None:
        """The port of the host. If not set, this value is `None`."""
        if self._port == -1:
            return None

        return self._port
    
    @port.setter
    def port(self, value: int|str|None):
        if value is None:
            self._port = -1
            return

        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError:
                error(f'Host: Invalid port "{value}"')
                return

        if value < 0 or value > 65535:
            error(f'Host: Invalid port "{value}"')
            return

        self._port = value

    @property
    def domain(self) -> str | None:
        """The domain of the host. Example: `www.google.com`
        
        If not set, this value is `None`.
        """
        if len(self._domain) == 0:
            return None

        return self._domain

    @domain.setter
    def domain(self, value: str|None):
        if value is None:
            self._domain = ''
            return

        self._domain = value

    @property
    def base_url(self) -> str | None:
        """The base URL of the host.
        
        This value is generated using `scheme`, `domain`/`ip` and `port`
        if they are available. Otherwise, it'll be `None`.

        When setting this, you can pass any full URL as value.
        If possible, `scheme`, `domain`/`ip` and `port` will be extracted from it.
        """
        if self.domain is None and self.ip is None:
            return None
        
        base = self.domain or self.ip
        if self.scheme is not None:
            base = f'{self.scheme}://{base}'
        if self.port is not None:
            base += f':{self.port}'

        return base
    
    @base_url.setter
    def base_url(self, value: str):
        match = URL_PATTERN.match(value)
        if match is None or match.group(4) is None:
            error(f'Host: Invalid URL "{value}"')
            return

        self.scheme = match.group(3)
        self.ip = match.group(5)
        self.domain = match.group(6)
        self.port = match.group(9)

    def __str__(self) -> str:
        return self.base_url or ''

    def __repr__(self) -> str:
        return (
            'Host('
            f'scheme={self.scheme}, '
            f'ip={self.ip}, '
            f'domain={self.domain}, '
            f'port={self.port}, '
            f'base_url={self.base_url}'
            ')'
        )

def _request(url: str, headers, cookies, session, auth,
             allow_redirects: bool, retry_on_connection_error: bool,
             retry_on_unexpected_status_code: bool, 
             expected_status_codes: list[int],
             data=None) -> requests.Response | None:

    request_type = 'GET' if data is None else 'POST'

    try:
        if data is None:
            r: requests.Response = session.get(
                url, headers=headers, cookies=cookies, auth=auth, 
                allow_redirects=allow_redirects, timeout=5, proxies=proxy
            )
        else:
            r: requests.Response = session.post(
                url, headers=headers, cookies=cookies, auth=auth, 
                allow_redirects=allow_redirects, timeout=5, proxies=proxy,
                data=data
            )

        if not r.status_code in expected_status_codes:
            description = http.client.responses.get(r.status_code, 'Unknown')
            error(f'{request_type}: Unexpected status code {r.status_code} ({description})')
            if retry_on_unexpected_status_code:
                raise RetryInterrupt

            return None

        return r
    except requests.exceptions.ConnectionError:
        error(f'{request_type}: Connection Error. Is the URL "{url}" valid?')
        if retry_on_connection_error:
            raise RetryInterrupt
    except requests.exceptions.Timeout:
        error(f'{request_type}: Timeout')
        if retry_on_connection_error:
            raise RetryInterrupt
    except requests.exceptions.RequestException as e:
        error(f'{request_type}: Exception: {e}')

    return None

def request_get(url: str, verbose: bool = True, 
                headers=None, cookies=None, session=requests, auth=None,
                allow_redirects: bool = True,
                expected_status_codes: list[int] = [200, 302],
                retry_on_connection_error: bool = True,
                retry_on_unexpected_status_code: bool = False,
                wait_time_between_retries: float = 0.4,
                max_retries: int = 10) -> requests.Response | None:
    """Performs a GET request to `url`, using (if provided) `headers`,
    `cookies`, `session`, `auth` and `allow_redirects`.
    
    If a connection error happens or the response status code isn't
    in `expected_status_codes` (and `retry_on_connection_error` or
    `retry_on_unexpected_status_code` are `True` respectively), retries
    the request up to `max_retries` times, waiting `wait_time_between_retries`
    (See `call_retriable_func` for more info)

    If `verbose` is `True`, prints the URL before performing the request.

    If the request is successful, returns its response. Otherwise, returns `None`.
    """
    if verbose:
        debug(f'GET: {url}')

    return call_retriable_func(
        _request, max_retries=max_retries, 
        wait_time_between_retries=wait_time_between_retries,
        url=url, headers=headers, cookies=cookies, session=session, auth=auth,
        allow_redirects=allow_redirects,
        retry_on_connection_error=retry_on_connection_error,
        retry_on_unexpected_status_code=retry_on_unexpected_status_code, 
        expected_status_codes=expected_status_codes
    )

def request_post(url: str, data, verbose: bool = True, 
                headers=None, cookies=None, session=requests, auth=None,
                allow_redirects: bool = True,
                expected_status_codes: list[int] = [200, 302],
                retry_on_connection_error: bool = True,
                retry_on_unexpected_status_code: bool = False,
                wait_time_between_retries: float = 0.4,
                max_retries: int = 10) -> requests.Response | None:
    """Performs a POST request to `url`, using (if provided) `headers`,
    `cookies`, `session`, `auth`, `allow_redirects` and `data`.
    
    If a connection error happens or the response status code isn't
    in `expected_status_codes` (and `retry_on_connection_error` or
    `retry_on_unexpected_status_code` are `True` respectively), retries
    the request up to `max_retries` times, waiting `wait_time_between_retries`
    (See `call_retriable_func` for more info)

    If `verbose` is `True`, prints the URL before performing the request.

    If the request is successful, returns its response. Otherwise, returns `None`.
    """
    if verbose:
        debug(f'POST: {url}')

    return call_retriable_func(
        _request, max_retries=max_retries, 
        wait_time_between_retries=wait_time_between_retries, url=url,
        data=data, headers=headers, cookies=cookies, session=session, auth=auth,
        allow_redirects=allow_redirects,
        retry_on_connection_error=retry_on_connection_error,
        retry_on_unexpected_status_code=retry_on_unexpected_status_code, 
        expected_status_codes=expected_status_codes
    )

def socket_connection(host, port, requests: list[str] = [], udp: bool = False, 
                      timeout: int = 3, decode_data: bool = True) -> list[str] | None:
    socket_kind = socket.SOCK_STREAM
    if udp:
        socket_kind = socket.SOCK_DGRAM

    s = socket.socket(socket.AF_INET, socket_kind)
    s.setblocking(False)
    s.settimeout(timeout)

    try:
        debug(f"SOCKET: Connecting to {host}:{port}")
        s.connect((host, int(port)))
    except Exception as e:
        error(f"Couldn't connect to {host}:{port}")
        error(str(e))
        return None
    
    result = []
    for request in requests:
        debug(f"SOCKET: Sending {request.rstrip()}")
        s.send(request.encode())

        full_data = bytes()
        while True:
            time.sleep(0.2)

            try:
                data = s.recv(512000) # 512 KB
            except socket.error as e:
                if e.args[0] == 'timed out':
                    warn("Timeout")
                    break
                else:
                    raise e
            
            if not data:
                break

            full_data += data
        
        if decode_data:
            full_data = full_data.decode('utf8', 'ignore')

        result.append(full_data)
    
    return result


import os
import platform
import shutil
import subprocess
import time

__all__ = [
    'IN_COLAB', 'OS', 'tool_exists', 'get_unique_file', 'run_and_print_output'
]

OS = platform.system()
# Check if we are being executed on Google Colab
try:
    import google.colab
    IN_COLAB = True
except ImportError:
    IN_COLAB = False

_tool_exists_cache = {}
def tool_exists(name: str) -> bool:
    """Check whether `name` is on PATH and marked as executable."""

    # Using a cache improves performance
    if _tool_exists_cache.get(name) is not None:
        return _tool_exists_cache[name]

    result = shutil.which(name) is not None
    _tool_exists_cache[name] = result
    return result

def get_unique_file(file_path: str) -> str:
    """Generates an unique file name for `file_path` and returns it.
    
    If `file_path` does not exist, returns it as-is.
    Otherwise, appends a number to it.
    """

    base_name = os.path.basename(file_path)
    extension = ''
    if '.' in base_name:
        # Make sure file_path is not something like ".gitignore"
        if len(base_name.split('.')[0]) > 0:
            extension = file_path[file_path.rfind('.'):]
            file_path = file_path[:file_path.rfind(extension)]

    unique_path = file_path + extension

    i = 1
    while os.path.exists(unique_path):
        unique_path = f"{file_path}_{i}{extension}"
        i += 1
    
    return unique_path

def run_and_print_output(print_func: callable, *args, **kwargs):
    """Calls `subprocess.Popen` with `args` and `kwargs`.
    
    Pipes both stdout and stderr, and every time the process writes
    a line to stdout, calls `print_func` with it.
    """

    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.STDOUT
    kwargs.setdefault('encoding', 'utf8')

    try:
        proc = subprocess.Popen(*args, **kwargs)
        while proc.poll() is None:
            line = proc.stdout.readline()
            if not line: 
                proc.wait()
                continue

            print_func(line)
    except KeyboardInterrupt:
        # Wait a little in order to prevent EOFError
        # when user presses CTRL + C
        time.sleep(0.1)
        raise KeyboardInterrupt


import base64
import urllib.parse
__all__ = [
    'base64_decode', 'base64_encode',
    'url_decode', 'url_encode'
]

def base64_decode(s: str) -> str:
    """Decodes a Base64-encoded string."""

    data_bytes = base64.b64decode(s.encode('utf8'))
    return data_bytes.decode('utf8')
def base64_encode(s: str) -> str:
    """Encodes a string with Base64."""

    data_bytes = base64.b64encode(s.encode('utf8'))
    return data_bytes.decode('utf8')

def url_decode(url: str) -> str:
    """Decodes (unquotes) `url`."""
    return urllib.parse.unquote(url)
def url_encode(url: str) -> str:
    """Encodes (quotes) `url`."""
    return urllib.parse.quote(url)


import copy
import os
import re

from .display import *
from .environment import IN_COLAB
__all__ = [
    'color_input', 'decorator_input', 'info_input', 'info2_input', 'success_input',
    'debug_input', 'warn_input', 'error_input', 'ask', 'input_float', 'input_int',
    'input_str', 'input_file', 'get_selection', 'conditional_get_selection',
    'get_selection_from_table'
]

def color_input(text: str) -> str:
    """Applies style to `text` and calls input()."""

    # Colab doesn't properly display text printed 
    # by input() when it's too long
    if IN_COLAB:
        color_print(text)
        return input()

    text += '[0]'
    text = apply_style(text)
    return input(text)

def decorator_input(decorator: str, text: str, 
                    end_text: str = '[0] [darkgray][Enter][0]') -> str:
    """Adds `decorator` at the beggining of `text` and 
    `end_text` at the end, then calls input()."""

    text, decorator = move_new_lines(text, decorator)
    return color_input(decorator + text + end_text)

def info_input(text: str) -> str:
    return decorator_input(decorators['info'], text)
def info2_input(text: str) -> str:
    return decorator_input(decorators['info2'], text)
def success_input(text: str) -> str:
    return decorator_input(decorators['success'], text)
def debug_input(text: str) -> str:
    return decorator_input(decorators['debug'], text, '[0] [Enter][0]')
def warn_input(text: str) -> str:
    return decorator_input(decorators['warn'], text)
def error_input(text: str) -> str:
    return decorator_input(decorators['error'], text)

def ask(msg: str, yes_default: bool = False) -> bool:
    """Asks the user to input Yes or No (Y/N). Returns `True` if the answer is Yes.
    
    If `yes_default` is `True`, yes will be selected if the user doesn't input anything.
    """

    decorator = '[bold][blue][?][0]'
    msg2 = '[Y/n]' if yes_default else '[y/N]'
    msg2 = '[cyan]' + msg2

    msg, decorator = move_new_lines(msg, decorator)
    selection = color_input(f'{decorator} {msg} {msg2} ').upper()

    if len(selection) == 0:
        return yes_default
    if selection != 'Y' and selection != 'N':
        clear_last_line()
        return ask(msg, yes_default)

    return selection == 'Y'

def input_str(msg: str, fallback = None,
              min_len: int|None = None, max_len: int|None = None) -> str:
    """Asks the user to input a string. Returns it.

    If `fallback` isn't `None`, it will be returned as-is if the user doesn't input anything.\n
    If `min_len` isn't `None`, the user won't be allowed to input a shorter string.\n
    If `max_len` isn't `None`, the user won't be allowed to input a longer string.
    """

    msg += ': '
    print()
    while 1:
        clear_last_line()
        # Print fallback value
        if fallback is not None:
            color_print((' ' * len(msg)) + f'[darkgray]{fallback}', end='\r')

        value = color_input(msg)
        if not IN_COLAB:
            # If fallback value is larger than the input, the gray
            # text will still be there so print the whole line again
            clear_last_line()
            if len(value) == 0 and fallback is not None:
                color_print(f'{msg}{fallback}')
            else:
                color_print(f'{msg}{value}')

        if len(value) == 0:
            if fallback is not None:
                return fallback
            continue
        if min_len is not None and len(value) < min_len:
            info_input(f"Input must have more than {min_len} chars")
            clear_last_line()
            continue
        if max_len is not None and len(value) > max_len:
            info_input(f"Input must have less than {max_len} chars")
            clear_last_line()
            continue
        return value

def _input_number(type: type, msg: str, fallback = None,
                  min = None, max = None, accepted_values = None):
    print()
    while 1:
        clear_last_line()
        value = input_str(msg, fallback)
        if value == fallback:
            # Return fallback as-is. Don't apply checks to it.
            return fallback

        try:
            result = type(value)
        except ValueError:
            continue

        if min is not None and result < min:
            info_input(f"Number can't be less than {min}")
            clear_last_line()
            continue
        if max is not None and result > max:
            info_input(f"Number can't be higher than {max}")
            clear_last_line()
            continue
        if accepted_values is not None and not result in accepted_values:
            info_input(f"Invalid value {accepted_values}")
            clear_last_line()
            continue

        return result
    
def input_float(msg: str, fallback = None,
                min: float|None = None, max: float|None = None,
                accepted_values: list[float]|None = None) -> float:
    """Asks the user to input a float. Returns it.
    
    If `fallback` isn't `None`, it will be returned as-is if the user doesn't input anything.\n
    If `min` isn't `None`, the user won't be allowed to input a lower number.\n
    If `max` isn't `None`, the user won't be allowed to input a higher number.\n
    If `accepted_values` isn't `None`, the user won't be allowed to input any value not inside it.
    """
    
    return _input_number(float, msg, fallback, min, max, accepted_values)

def input_int(msg: str, fallback = None, 
              min: int|None = None, max: int|None = None,
              accepted_values: list[int]|None = None) -> int:
    """Asks the user to input an integer. Returns it.

    If `fallback` isn't `None`, it will be returned as-is if the user doesn't input anything.\n
    If `min` isn't `None`, the user won't be allowed to input a lower number.\n
    If `max` isn't `None`, the user won't be allowed to input a higher number.\n
    If `accepted_values` isn't `None`, the user won't be allowed to input any value not inside it.
    """

    return _input_number(int, msg, fallback, min, max, accepted_values)

def input_file(msg: str, fallback: str|None = None) -> str:
    """Asks the user to input a valid path of a file. Returns it.
    
    If `fallback` isn't `None`, it will be returned if the user
    doesn't input anything AND `fallback` is a valid file path.
    """

    while 1:
        file = input_str(msg, fallback)
        # Strip quotes, this is useful because on Windows the
        # file explorer has an option to copy the path of a folder/file,
        # and it adds quotes to it
        file = file.strip('"')

        if os.path.isfile(file):
            break
        clear_last_line()

    return file

def get_selection(*options, input_msg: str = 'Please select an option',
                  commands: dict[re.Pattern, callable] = {},
                  print_func: callable = print,
                  invert_print_order: bool = False) -> str:
    """Shows the user a list with all the given options.
       Asks them to select one and returns it.
       
       If `commands` is supplied, all of its patterns will be tested against the user
       input, and if one of them matches, its function will be executed.
       
       `print_func` can be used to customize the list style by replacing the print
       function with a custom one.

       If `invert_print_order` is `True`, the options will be printed in reverse order.
    """

    if not invert_print_order:
        for i, option in enumerate(options):
            print_func(f'{i + 1}) {option}')
    else:
        for i, option in reversed(list(enumerate(options))):
            print_func(f'{i + 1}) {option}')

    print()
    while 1:
        selection_str = color_input(f'[darkgray]{input_msg}: ')

        # Detect commands and execute them
        for pattern, func in commands.items():
            match = re.match(pattern, selection_str)
            if match is not None:
                func(match)
                break

        try:
            selection = int(selection_str)
        except ValueError:
            clear_last_line()
            continue

        if selection > len(options) or selection < 1:
            clear_last_line()
            continue
            
        break

    return options[selection - 1]

def conditional_get_selection(options: dict[str, bool],
                              input_msg = 'Please select an option',
                              commands: dict[re.Pattern, callable] = {},
                              print_func: callable = print,
                              invert_print_order: bool = False) -> str:
    """Shows the user a list with the keys from `options` whose value are `True`. 
       Asks them to select one and returns it.
       
       If `commands` is supplied, all of its patterns will be tested against the user
       input, and if one of them matches, its function will be executed.

       `print_func` can be used to customize the list style by replacing the print
       function with a custom one.

       If `invert_print_order` is `True`, the options will be printed in reverse order.
    """

    options = [key for key, value in options.items() if value]
    return get_selection(*options, input_msg=input_msg, commands=commands, 
                         print_func=print_func, invert_print_order=invert_print_order)

def get_selection_from_table(columns: dict[str, int], options: list[list[str]],
                             input_msg: str = 'Please select an option',
                             commands: dict[re.Pattern, callable] = {},
                             invert_print_order: bool = False) -> int:
    """Just like `get_selection()`, but uses `print_table()` to print the options,
    and returns the index of the selected one."""

    columns = {'#': 4} | columns

    # We need to make a shallow copy of the options and
    # row lists so the option index is not added to the original one
    options = copy.copy(options)
    for i, row in enumerate(options):
        row = copy.copy(row)
        row.insert(0, f'{i + 1})')
        options[i] = row

    print_table(columns, options, invert_print_order=invert_print_order)
    selection = get_selection(*options, print_func=lambda *_: None,
                              input_msg=input_msg, commands=commands)

    for i, row in enumerate(options):
        if selection == row:
            return i

    raise Exception('Selection index not found')


import math
import os
import subprocess
import sys

from .environment import *
__all__ = [
    'title', 'style', 'get_terminal_width', 'apply_style', 'remove_style',
    'color_print', 'move_new_lines', 'decorator_print', 'info', 'info2',
    'success', 'debug', 'warn', 'error', 'clear', 'clear_last_line', 
    'print_separator', 'print_title', 'print_table', 'decorators'
]

title = 'PLACEHOLDER'
style = {
    '0':            '\033[0m',
    'red':          '\033[31m',
    'green':        '\033[32m',
    'yellow':       '\033[33m',
    'blue':         '\033[34m',
    'purple':       '\033[35m',
    'cyan':         '\033[36m',
    'darkgray':     '\033[90m',
    'white':        '\033[97m',

    'bold':         '\033[1m',
#   'dim':          '\033[2m', # Doesn't work in Kali Linux
    'italic':       '\033[3m',
    'underline':    '\033[4m',
#   'blink':        '\033[5m', # Doesn't work in Termux
    'invert':       '\033[7m'
}
decorators = {
    'info':    '[bold][cyan][*][0][cyan] ',
    'info2':   '[bold][purple][*][0][purple] ',
    'success': '[bold][green][+][0][green] ',
    'debug':   '[bold][green][*][0][darkgray] ',
    'warn':    '[bold][yellow][!][0][yellow] ',
    'error':   '[bold][red][!][0][red] '
}

def get_terminal_width(fallback: int = 50) -> int:
    """Returns the width (columns) of the terminal.
    
    If we are running in Google Colab, returns `fallback`.
    """

    # Colab doesn't support getting the width of the output box
    if IN_COLAB:
        return fallback
    
    return os.get_terminal_size().columns

def apply_style(text: str) -> str:
    """Applies style to `text` and returns it."""

    for key, value in style.items():
        text = text.replace(f'[{key}]', value)
    return text

def remove_style(text: str) -> str:
    """Removes all style from `text` and returns it."""

    for key in style.keys():
        text = text.replace(f'[{key}]', '')
    return text

def color_print(text: str, output=sys.stdout, end: str = '\n'):
    """Applies style to `text`, appends `end` (newline by default)
    and prints it to `output`.
    """

    text += f'[0]{end}'
    text = apply_style(text)
    output.write(text)
    output.flush()

def move_new_lines(src: str, dst: str) -> tuple[str, str]:
    """Moves all newlines at the beginning of `src` to `dst`."""

    while src.startswith('\n'):
        src = src[1:]
        dst = '\n' + dst
    
    return (src, dst)

def decorator_print(decorator: str, text: str):
    """Prints `text` adding `decorator` at the beginning."""

    text, decorator = move_new_lines(text, decorator)
    color_print(decorator + text)

def info(text: str):
    decorator_print(decorators['info'], text)
def info2(text: str):
    decorator_print(decorators['info2'], text)
def success(text: str):
    decorator_print(decorators['success'], text)
def debug(text: str):
    decorator_print(decorators['debug'], text)
def warn(text: str):
    decorator_print(decorators['warn'], text)
def error(text: str):
    decorator_print(decorators['error'], text)

def clear():
    """Executes clear (Linux) or cls (Windows)."""

    # clear() doesn't work in Colab
    if IN_COLAB:
        return

    if OS == 'Windows': 
        subprocess.run('cls', shell=True)
    else: 
        subprocess.run('clear', shell=True)

def clear_last_line():
    """Clears the last line in the console, overwriting it."""

    # This doesn't work in Colab
    if IN_COLAB:
        return

    whitespace = ' ' * get_terminal_width()
    print(f'\033[A{whitespace}\033[A')

def print_separator(char: str, style: str = ''):
    """Prints `char` until it fills the terminal horizontally."""

    width = get_terminal_width()
    separator = char * width
    color_print(style + separator.center(width))

def print_title(subtitle: str = '',
                separator_char: str = '=', style: str = '[cyan]'):
    """Prints `TITLE` and `subtitle` between two separators."""

    clear()
    if len(subtitle) > 0:
        subtitle = ' | ' + subtitle

    print_separator(separator_char, style)
    color_print(f'[bold]{style}' + (title + subtitle).center(get_terminal_width()))
    print_separator(separator_char, style)

def print_table(columns: dict[str, int], data: list[list[str]],
                invert_print_order: bool = False):
    """Prints a table.
    
    `columns`: `Key`: Column name. `Value`: Column length.\n
    If a column's length is `0`, it'll take the biggest possible length.
    If it's `-1`, the column and its data will be ignored.

    `data`: a list of lists, where each inner list represents a row of data
    and must have the same length than `columns`.

    If `invert_print_order` is `True`, the table will be printed from bottom to top.
    """

    terminal_columns = get_terminal_width()

    dynamic_columns_amount = 0
    fixed_columns_size = 0
    for column_len in columns.values():
        if column_len == 0:
            dynamic_columns_amount += 1
        elif column_len > 0:
            fixed_columns_size += column_len

    dynamic_columns_len = 0
    if dynamic_columns_amount > 0:
        dynamic_columns_len = math.floor(
            (terminal_columns - fixed_columns_size) / dynamic_columns_amount
        )

    print()

    # Print header
    header = '[bold][white]'
    for name, length in columns.items():
        if length == 0:
            length = dynamic_columns_len
        elif length == -1:
            continue

        # Truncate name if needed
        if len(name) > length:
            name = name[:length - 3] + '...'

        header += '{:<{}}'.format(name, length)
    color_print(header)

    if invert_print_order:
        data.reverse()

    # Print data
    for row in data:
        row_str = ''

        for i, (name, length) in enumerate(columns.items()):
                if length == 0:
                    length = dynamic_columns_len
                elif length == -1:
                    continue

                value = remove_style(row[i])
                style = row[i].replace(value, '')

                # Truncate value if needed
                if len(value) > length:
                    value = value[:length - 3] + '...'

                row_str += style + '{:<{}}'.format(value, length)

        color_print(row_str)